import re
from z3 import *

TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255

def safe_decode(hex_encoded_string):

    if (hex_encoded_string.startswith("0x")):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)

def to_signed(i):
    return i if i < TT255 else i - TT256

def get_instruction_index(instruction_list, address):

	index = 0

	for instr in instruction_list:
		if instr['address'] == address:
			return index
		index += 1

	return None

def get_trace_line(instr, state):

	stack = str(state.stack[::-1])

	stack = re.sub("(\d+)",	lambda m: hex(int(m.group(1))), stack)
	stack = re.sub("\n", "", stack)

	return str(instr['address']) + " " + instr['opcode'] + "\tSTACK: " + stack

def pop_bitvec(state):
    # pop one element from stack, converting boolean expression to bitvector

    item = state.stack.pop()

    if type(item) == BoolRef:
        return If(item, BitVecVal(1, 256), BitVecVal(0, 256))
    else:
        return item

def solve_path(svm, path, caller = None, owner = None, owner_storage_index = None):

    s = Solver() 

    if(caller is not None):
        s.add(svm.env['caller'] == caller)
    if(owner is not None):
        s.add(svm.storage[owner_storage_index] == owner)

    for edge in path:
        if edge.condition is not None:
            s.add(edge.condition)

    if (s.check() == sat):
        return s.model()

    else:
        return None



import graphviz as gv
from z3 import *
import re

styles = {
    'graph': {
        'overlap': 'false',
        'fontsize': '12',
        'fontcolor': 'white',
        'bgcolor': '#333333',
        'concentrate':'true',
    },
    'nodes': {
        'fontname': 'Helvetica',
        'fontsize': '12',
        'shape': 'box',
        'fontcolor': 'white',
        'color': 'white',
        'style': 'filled',
        'fillcolor': '#009966',
    },
    'edges': {
        'style': 'dashed',
        'dir': 'forward',
        'color': 'white',
        'arrowhead': 'normal',
        'fontname': 'Courier',
        'fontsize': '12',
        'fontcolor': 'white',
    }
}

def apply_styles(graph, styles):
    graph.graph_attr.update(
        ('graph' in styles and styles['graph']) or {}
    )
    graph.node_attr.update(
        ('nodes' in styles and styles['nodes']) or {}
    )
    graph.edge_attr.update(
        ('edges' in styles and styles['edges']) or {}
    )
    return graph


def generate_callgraph(svm, file):

    graph = gv.Graph(format='svg')

    for key in svm.nodes:
        node_text = ""

        for instruction in svm.nodes[key].instruction_list:
            node_text += str(instruction['address']) + " " + instruction['opcode']
            if instruction['opcode'].startswith("PUSH"):
                node_text += " " + instruction['argument']

            node_text += "\l"

        graph.node(str(key), node_text)

    for edge in svm.edges:

        if (edge.condition is None):
            simplified = ""
        else:

            try:
                simplified = str(simplify(edge.condition))
            except Z3Exception:
                simplified = str(edge.condition)
            
            simplified = re.sub("([\d+)",  lambda m: hex(int(m.group(1))), simplified)
            simplified = re.sub("[0]{8}[0]+", "0000(...)", simplified)
            simplified = re.sub("[f]{8}[f]+", "ffff(...)", simplified)

        graph.edge(str(edge.node_from),str(edge.node_to), simplified)

    graph = apply_styles(graph, styles)

    graph.render(file)




from enum import Enum
from laser.ethereum import utils
from z3 import *
import copy
import logging
from random import randint


TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255

MAX_DEPTH = 16


class SVMError(Exception):
    pass


class JumpType(Enum):
    CONDITIONAL = 1
    UNCONDITIONAL = 2


class State(): 

    def __init__(self):

        self.calldata = {}
        self.memory = {}
        self.stack = []
        self.pc = 0

    def calldata_alloc(self, offset):
        data = BitVec("calldata_" + str(offset), 256)
        return data


class Node:

    def __init__(self, start_addr=0):
        self.start_addr = start_addr
        self.instruction_list = []

    def __str__(self):
        return str(self.as_dict())
        
    def as_dict(self):

        code = ""

        for instruction in self.instruction_list:
            code += str(instruction['address']) + " " + instruction['opcode']
            if instruction['opcode'].startswith("PUSH"):
                code += " " + instruction['argument']

            code += "\\n"

        return {'id': self.start_addr, 'code': code}


class Edge:
    
    def __init__(self, node_from, node_to, edge_type=JumpType.UNCONDITIONAL, condition=None):

        self.node_from = node_from
        self.node_to = node_to
        self.type = edge_type
        self.condition = condition

    def __str__(self):
        return str(self.as_dict())
        
    def as_dict(self):

        return {'from': self.node_from, 'to': self.node_to}

class SVM:

    def __init__(self, disassembly, max_depth=MAX_DEPTH, branch_at_jumpi = False):
        self.disassembly = disassembly
        self.nodes = {}
        self.addr_visited = []
        self.edges = []
        self.paths = {}
        self.send_eth_locs = []
        self.reentrancy_funcs = []
        self.sstor_node_lists = {}
        self.storage = {}
        self.function_state = {}
        self.trace = ""
        self.max_depth = max_depth
        self.branch_at_jumpi = branch_at_jumpi

        self.env = {}

        self.env['callvalue'] = BitVec("callvalue", 256)
        self.env['caller'] = BitVec("caller", 256)
        self.env['origin'] = BitVec("origin", 256)
        self.env['address_to'] = BitVec("address_to", 256) 


    def depth_first_search(self, this_node, node_to, path, paths, depth, nodes_visited):

        if (depth > MAX_DEPTH):
            return

        if this_node == node_to:
            paths.append(path)
            return

        nodes_visited.append(this_node)

        edges_out = []

        for edge in self.edges:

            if edge.node_from == this_node and edge.node_to not in nodes_visited:
                edges_out.append(edge)

        for edge in edges_out:

            new_path = copy.deepcopy(path)
            new_path.append(edge)
            self.depth_first_search(edge.node_to, node_to, new_path, paths, depth + 1, nodes_visited)


    def find_paths(self, node_to):

        paths = []
        nodes_visited = []

        self.depth_first_search(0, node_to, [], paths, 0, nodes_visited)

        return paths


    def sym_exec(self):

        logging.debug("Starting SVM execution")

        self.nodes[0] = self._sym_exec(State(), 0)

        logging.info(str(len(self.nodes)) + " nodes, " + str(len(self.edges)) + " edges")

        logging.info("Resolving paths")

        for key in self.nodes:

            paths = self.find_paths(key)

            logging.debug("Found " + str(len(paths)) + " paths to node " + str(key))

            self.paths[key] = paths


    def _sym_exec(self, state, depth):
    
        start_addr = self.disassembly.instruction_list[state.pc]['address']
        node = Node(start_addr)

        logging.debug("- Entering new block, index = " + str(state.pc) + ", address = " + str(start_addr) + ", depth = " + str(depth))

        if start_addr == 0:
            self.function_state['current_func'] = "prologue"
            self.function_state['current_func_addr'] = start_addr
            self.function_state['sstore_called'] = False            

        if start_addr in self.disassembly.addr_to_func:
            # Start of a function

            function_name = self.disassembly.addr_to_func[start_addr]

            self.function_state['current_func'] = function_name
            self.function_state['sstore_called'] = False

            logging.info("- Entering function " + function_name)

            node.instruction_list.append({'opcode': function_name, 'address': self.disassembly.instruction_list[state.pc]['address']})

            state.pc += 1

        halt = False

        instr = self.disassembly.instruction_list[state.pc]

        while not halt:

            instr = self.disassembly.instruction_list[state.pc]
            node.instruction_list.append(instr)

            op = instr['opcode']

            logging.debug(utils.get_trace_line(instr, state))

            state.pc += 1

            # stack ops
            
            if op.startswith("PUSH"):
                value = BitVecVal(int(instr['argument'][2:], 16), 256)
                state.stack.append(value)

            elif op.startswith('DUP'):
                depth = int(op[3:])
                state.stack.append(state.stack[-depth])

            elif op.startswith('SWAP'):
                depth = int(op[4:])
                temp = state.stack[-depth - 1]
                state.stack[-depth - 1] = state.stack[-1]
                state.stack[-1] = temp

            elif op == 'POP':
                state.stack.pop()

            # Bitwise ops

            elif op == 'AND':
                op1 = state.stack.pop() 
                op2 = state.stack.pop()

                state.stack.append(op1 & op2)

            elif op == 'OR':
                op1 = state.stack.pop()
                op2 = state.stack.pop()

                if (type(op1) == BoolRef):
                    op1 = If(op1, BitVecVal(1,256), BitVecVal(0,256))

                if (type(op2) == BoolRef):
                    op2 = If(op2, BitVecVal(1,256), BitVecVal(0,256))

                state.stack.append(op1 | op2)

            elif op == 'XOR':
                state.stack.append(state.stack.pop() ^ state.stack.pop())

            elif op == 'NOT':
                state.stack.append(TT256M1 - state.stack.pop())

            elif op == 'BYTE':
                s0, s1 = state.stack.pop(), state.stack.pop()

                if s0 >= 32:
                    state.stack.append(0)
                else:
                    state.stack.append((s1 // 256 ** (31 - s0)) % 256)

            # Arithmetics

            elif op == "ADD":
                state.stack.append((utils.pop_bitvec(state) + utils.pop_bitvec(state)))

            elif op == "SUB":
                state.stack.append((utils.pop_bitvec(state) - utils.pop_bitvec(state)))

            elif op == 'MUL':
                state.stack.append(utils.pop_bitvec(state) * utils.pop_bitvec(state))

            elif op == 'DIV':
                s0, s1 = utils.pop_bitvec(state), utils.pop_bitvec(state)
                state.stack.append(UDiv(s0, s1))

            elif op == 'MOD':
                s0, s1 = utils.pop_bitvec(state), utils.pop_bitvec(state)
                state.stack.append(0 if s1 == 0 else s0 % s1)

            elif op == 'SDIV':
                s0, s1 = utils.to_signed(utils.pop_bitvec(state)), utils.to_signed(utils.pop_bitvec(state))
                state.stack.append(0 if s1 == 0 else (abs(s0) // abs(s1) *
                                              (-1 if s0 * s1 < 0 else 1)) & TT256M1)

            elif op == 'SMOD':
                s0, s1 = utils.to_signed(utils.pop_bitvec(state)), utils.to_signed(utils.pop_bitvec(state))
                state.stack.append(0 if s1 == 0 else (abs(s0) % abs(s1) *
                                              (-1 if s0 < 0 else 1)) & TT256M1)

            elif op == 'ADDMOD':
                s0, s1, s2 = utils.pop_bitvec(state), utils.pop_bitvec(state), utils.pop_bitvec(state)
                state.stack.append((s0 + s1) % s2 if s2 else 0)

            elif op == 'MULMOD':
                s0, s1, s2 = utils.pop_bitvec(state), utils.pop_bitvec(state), utils.pop_bitvec(state)
                state.stack.append((s0 * s1) % s2 if s2 else 0)

            elif op == 'EXP':
                # we only implement 2 ** x
                base, exponent = state.stack.pop(), state.stack.pop()

                if (base.as_long() == 2):
                    if exponent == 0:
                        state.stack.append(BitVecVal(1, 256))
                    else:
                        state.stack.append(base << (exponent - 1))

                else:
                    state.stack.append(base)

            elif op == 'SIGNEXTEND':
                s0, s1 = state.stack.pop(), state.stack.pop()
                if s0 <= 31:
                    testbit = s0 * 8 + 7
                    if s1 & (1 << testbit):
                        state.stack.append(s1 | (TT256 - (1 << testbit)))
                    else:
                        state.stack.append(s1 & ((1 << testbit) - 1))
                else:
                    state.stack.append(s1)

            # Comparisons

            elif op == 'LT':

                exp = ULT(utils.pop_bitvec(state), utils.pop_bitvec(state))
                state.stack.append(exp)

            elif op == 'GT':

                exp = UGT(utils.pop_bitvec(state), utils.pop_bitvec(state))
                state.stack.append(exp)

            elif op == 'SLT':

                exp = SLT(utils.pop_bitvec(state) < utils.pop_bitvec(state))
                state.stack.append(exp)

            elif op == 'SGT':

                exp = SGT(utils.pop_bitvec(state) > utils.pop_bitvec(state))
                state.stack.append(exp)

            elif op == 'EQ':

                exp = state.stack.pop() == state.stack.pop()
                state.stack.append(exp)

            elif op == 'ISZERO':

                val = state.stack.pop()

                if (type(val) == BoolRef):
                   exp = val == False
                else:   
                   exp = val == 0

                state.stack.append(exp)

             # Call data

            elif op == 'CALLVALUE':
                state.stack.append(self.env['callvalue'])

            elif op == 'CALLDATALOAD':
                offset = state.stack.pop()
                state.stack.append(state.calldata_alloc(offset))

            elif op == 'CALLDATASIZE':
                state.stack.append(BitVec("calldatasize", 256))

            elif op == 'CALLDATACOPY':
                mstart, dstart, size = state.stack.pop(), state.stack.pop(), state.stack.pop()

            # Control flow

            elif op == 'STOP':
                return node

            # Environment

            elif op == 'ADDRESS':
                state.stack.append(self.env['address_to'])

            elif op == 'BALANCE':
                addr = state.stack.pop()
                state.stack.append(BitVec("balance_at_" + str(addr), 256))

            elif op == 'ORIGIN':
                state.stack.append(self.env['origin'])

            elif op == 'CALLER':
                state.stack.append(self.env['caller'])

            elif op == 'CODESIZE':
                state.stack.append(len(self.disassembly.instruction_list))

            if op == 'SHA3':
                s0, s1 = utils.pop_bitvec(state), utils.pop_bitvec(state)
                state.stack.append(BitVec("sha_hash", 256))

            elif op == 'GASPRICE':
                state.stack.append(0)

            elif op == 'CODECOPY':
                start, s1, size = state.stack.pop(), state.stack.pop(), state.stack.pop()
                # Not implemented

            elif op == 'EXTCODESIZE':
                addr = state.stack.pop()
                state.stack.append(BitVec("extcodesize", 256))

            elif op == 'EXTCODECOPY':
                addr = state.stack.pop()
                start, s2, size = state.stack.pop(), state.stack.pop(), state.stack.pop()
                # Not implemented

            elif op == 'BLOCKHASH':
                state.stack.append(BitVec("blockhash", 256))

            elif op == 'COINBASE':
                state.stack.append(BitVec("coinbase", 256))

            elif op == 'TIMESTAMP':
                state.stack.append(BitVec("timestamp", 256))

            elif op == 'NUMBER':
                state.stack.append(BitVec("block_number", 256))

            elif op == 'DIFFICULTY':
                state.stack.append(BitVec("block_difficulty", 256))

            elif op == 'GASLIMIT':
                state.stack.append(BitVec("block_gaslimit", 256))

            elif op == 'MLOAD':
                offset = state.stack.pop()

                try:
                    data = state.memory[offset]
                except KeyError:
                    state.memory[offset] = BitVec("mem_" + str(offset), 256)
                    data = state.memory[offset]

                state.stack.append(data)

            elif op == 'MSTORE':
                offset, value = state.stack.pop(), state.stack.pop()

                state.memory[offset] = value

            elif op == 'MSTORE8':
                offset, value = state.stack.pop(), state.stack.pop()

                state.memory[offset] = value % 256

            elif op == 'SLOAD':
                offset = state.stack.pop()
                logging.debug("Storage access at offset " + str(offset))

                try:
                    data = self.storage[offset]
                except KeyError:
                    data = BitVec("storage_" + str(offset), 256)
                    self.storage[offset] = data

                state.stack.append(data)

            elif op == 'SSTORE':
                offset, value = state.stack.pop(), state.stack.pop()

                self.function_state['sstore_called'] = True

                if type(offset) == BitVecRef:
                    logging.debug("Not supported: SSTORE to hash offset")
                else:
                    i = offset.as_long()

                    logging.debug("Write to storage[" + str(offset) + "] at node " + str(start_addr))

                    try:
                        self.sstor_node_lists[i].append(start_addr)
                    except KeyError:
                        self.sstor_node_lists[i] = [start_addr]

                    try:
                        self.storage[i]
                    except KeyError:
                        self.storage[i] = BitVec("storage_" + str(offset), 256)

            elif op == 'JUMP':

                jump_addr = state.stack.pop()

                if (type(jump_addr) == BitVecRef):
                    logging.debug("Invalid jump argument: JUMP <bitvector> at " + str(self.disassembly.instruction_list[state.pc]['address']))

                    return node

                if (depth < self.max_depth):

                    jump_addr = jump_addr.as_long()

                    logging.debug("JUMP to " + str(jump_addr))

                    i = utils.get_instruction_index(self.disassembly.instruction_list, jump_addr)

                    if self.disassembly.instruction_list[i]['opcode'] == "JUMPDEST":
                        logging.debug("Current nodes: " + str(self.nodes))

                        if jump_addr not in self.addr_visited:
                            self.addr_visited.append(jump_addr)
                            new_state = copy.deepcopy(state)
                            new_state.pc = i
                            new_node = self._sym_exec(new_state, depth + 1)
                            self.nodes[jump_addr] = new_node

                        self.edges.append(Edge(node.start_addr, jump_addr, JumpType.UNCONDITIONAL))
                    else:
                        self.trace += "Skipping invalid jump destination"

                return node

            elif op == 'JUMPI':
                jump_addr, condition = state.stack.pop().as_long(), state.stack.pop()

                if (depth < self.max_depth):

                    i = utils.get_instruction_index(self.disassembly.instruction_list, jump_addr)

                    logging.debug("JUMPI to " + str(jump_addr))

                    if not i:
                        logging.debug("Invalid jump destination: " + str(jump_addr))
                    else:
                        instr = self.disassembly.instruction_list[i]

                        # Add new node for condition == True

                        if instr['opcode'] != "JUMPDEST":
                            logging.debug("Invalid jump destination: " + str(jump_addr))
                        else:
                            if jump_addr not in self.addr_visited:

                                self.addr_visited.append(jump_addr)

                                new_state = copy.deepcopy(state)

                                new_state.pc = i

                                new_node = self._sym_exec(new_state, depth + 1)
                                self.nodes[jump_addr] = new_node

                            logging.debug("Adding edge with condition: " + str(condition))

                            self.edges.append(Edge(node.start_addr, jump_addr, JumpType.CONDITIONAL, condition))

                        if (self.branch_at_jumpi):

                            # Add new node for condition == False

                            new_state = copy.deepcopy(state)

                            new_node = self._sym_exec(new_state, depth)

                            start_addr = self.disassembly.instruction_list[state.pc]['address']
                            self.nodes[start_addr] = new_node

                            self.edges.append(Edge(node.start_addr, start_addr, JumpType.CONDITIONAL, Not(condition)))

                            return node


            elif op == 'PC':
                state.stack.append(state.pc - 1)

            elif op == 'MSIZE':
                state.stack.append(BitVec("msize", 256))

            elif op == 'GAS':
                state.stack.append(10000000)

            elif op.startswith('LOG'):
                depth = int(op[3:])
                mstart, msz = state.stack.pop(), state.stack.pop()
                topics = [state.stack.pop() for x in range(depth)]
                # Not supported

            elif op == 'CREATE':
                value, mstart, msz = state.stack.pop(), state.stack.pop(), state.stack.pop()
                # Not supported
                state.stack.append(0)

            elif op == 'CALL':
                gas, to, value, meminstart, meminsz, memoutstart, memoutsz = \
                state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop()

                send_eth = False

                if (type(value) is not BitVecNumRef):
                    send_eth = True
                elif value.as_long() > 0:
                    send_eth = True

                if (send_eth):
                    logging.debug("CALL with non-zero value: " + str(value))
                    self.send_eth_locs.append({'address': start_addr, 'function_name': self.function_state['current_func']})

                    if not self.function_state['sstore_called']:
                        logging.info("Possible reentrancy at " + self.function_state['current_func'])
                        self.reentrancy_funcs.append(self.function_state['current_func_addr'])

                ret = BitVec("retval_" + str(self.disassembly.instruction_list[state.pc]['address']) + "_" + str(randint(0, 1000)), 256)

                state.stack.append(ret)

            elif op == 'CALLCODE':
                gas, to, value, meminstart, meminsz, memoutstart, memoutsz = \
                state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop(), state.stack.pop()
                # Not supported          
                state.stack.append(BitVecVal(0, 256))

            elif op == 'RETURN':
                offset, length = state.stack.pop(), state.stack.pop()
                # Not supported        
                # logging.debug("Returning from block " +  str(start_addr))
                return node

            elif op == 'SUICIDE':
                # logging.debug("Returning from block " +  str(start_addr))
                return node

            elif op == 'REVERT':
                # logging.debug("Returning from block " +  str(start_addr))
                return node

            elif op == 'INVALID':
                # logging.debug("Returning from block " +  str(start_addr))
                return node




from laser.ethereum import svm
from z3 import Z3Exception, simplify
import re


graph_html = '''<html>
 <head>
  <style type="text/css">
   #mynetwork {
    background-color: #000000;
   }

   body {
    background-color: #000000;
    color: #ffffff;
   }
  </style>
  <link href="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.css" rel="stylesheet" type="text/css" />
  <script src="https://cdnjs.cloudflare.com/ajax/libs/vis/4.21.0/vis.min.js"></script>
  <script>

    var options = {
      autoResize: true,
      height: '100%',
      width: '100%',
      manipulation: false,
      height: '90%',
      layout: {
        randomSeed: undefined,
        improvedLayout:true,
        hierarchical: {
          enabled:true,
          levelSeparation: 450,
          nodeSpacing: 200,
          treeSpacing: 100,
          blockShifting: true,
          edgeMinimization: true,
          parentCentralization: false,
          direction: 'LR',        // UD, DU, LR, RL
          sortMethod: 'directed'   // hubsize, directed
        }
      },
      nodes:{
        borderWidth: 1,
        borderWidthSelected: 2,
        chosen: true,
        shape: 'box',
        font: {
          align: 'left',
          color: '#FFFFFF',
        },
        color: {
          border: '#26996f',
          background: '#1f7e5b',
          highlight: {
            border: '#26996f',
            background: '#28a16f'
          }  
        }
      },
      edges:{
        font: {
          color: '#ffffff',
          size: 12, // px
          face: 'arial',
          background: 'none',
          strokeWidth: 0, // px
          strokeColor: '#ffffff',
          align: 'horizontal',
          multi: false,
          vadjust: 0,
        }
      },

      physics:{
        enabled: [ENABLE_PHYSICS],
      }   
    
  }

  [JS]

  </script>
 </head>
<body>
<p>Mythril / Ethereum LASER Symbolic VM</p>
<p><div id="mynetwork"></div><br /></p>
<script type="text/javascript">
var container = document.getElementById('mynetwork');
var data = {'nodes': nodes, 'edges': edges}
var gph = new vis.Network(container, data, options);
</script>
</body>
</html>
'''


def serialize(_svm):

    nodes = []
    edges = []

    for n in _svm.nodes:

        code =  _svm.nodes[n].as_dict()['code']

        code = re.sub("([0-9a-f]{8})[0-9a-f]+",  lambda m: m.group(1) + "(...)", code)

        nodes.append("{id: " + str(_svm.nodes[n].as_dict()['id']) + ", size: 150, 'label': '" + code + "'}")

    for edge in _svm.edges:

      if (edge.condition is None):
          label = ""
      else:

          try:
            label = str(simplify(edge.condition))
          except Z3Exception:
            label = str(edge.condition).replace("\n", "")
      
      label = re.sub("[^_]([[\d]{2}\d+)",  lambda m: hex(int(m.group(1))), label)
      code = re.sub("([0-9a-f]{8})[0-9a-f]+",  lambda m: m.group(1) + "(...)", code)

      edges.append("{from: " + str(edge.as_dict()['from']) + ', to: ' + str(edge.as_dict()['to']) + ", 'arrows': 'to', 'label': '" + label + "', 'smooth': {'type': 'cubicBezier'}}")

    return "var nodes = [\n" + ",\n".join(nodes) + "\n];\nvar edges = [\n" + ",\n".join(edges) + "\n];"



def generate_callgraph(disassembly, physics):

    _svm = svm.SVM(disassembly)

    _svm.sym_exec()

    html = graph_html.replace("[JS]", serialize(_svm))
    html = html.replace("[ENABLE_PHYSICS]", str(physics).lower())

    return html



from mythril.ether import asm,util
import os
import json


class Block:

    def __init__(self, id, code_index, funcname):
        self.id = id
        self.code_index = code_index
        self.funcname = funcname
        self.instruction_list = []

    def update_length(self, num_instructions):
        self.length = num_instructions
        self.start_addr = self.instruction_list[0]['address']
        self.end_addr = self.instruction_list[-1]['address']

    def get_easm(self):

        easm = str(self.instruction_list[0]['address']) + " " + self.funcname + "\n"
        easm += asm.instruction_list_to_easm(self.instruction_list[1:])

        return easm


class Disassembly:

    def __init__(self, code):
        self.instruction_list = asm.disassemble(util.safe_decode(code))
        self.blocks = []
        self.xrefs = []
        self.func_to_addr = {}
        self.addr_to_func = {}

        # Parse jump table & resolve function names

        script_dir = os.path.dirname(os.path.realpath(__file__))
        signature_file = os.path.join(script_dir, 'signatures.json')

        with open(signature_file) as f:
            signatures = json.load(f)

        jmptable_indices = asm.find_opcode_sequence(["PUSH4", "EQ"], self.instruction_list)

        for i in jmptable_indices:
            func_hash = self.instruction_list[i]['argument']
            try:
                func_name = signatures[func_hash]
            except KeyError:
                func_name = "_function_" + func_hash

            try:
                offset = self.instruction_list[i+2]['argument']
                jump_target = int(offset, 16)

                self.func_to_addr[func_name] = jump_target
                self.addr_to_func[jump_target] = func_name
            except:
                continue

        # Parse instructions into basic blocks

        current_block = Block(0, 0, "PROLOGUE")

        index = 0
        blocklen = 0
        blocknumber = 1

        for instruction in self.instruction_list:

            if (instruction['opcode'] == "JUMPDEST"):

                try:
                    func_name = "- FUNCTION " + self.addr_to_func[instruction['address']] + " -"
                except KeyError:
                    func_name = "- JUMPDEST_UNK -"
                
                current_block.update_length(blocklen)
                self.blocks.append(current_block)
                current_block = Block(blocknumber, index, func_name)
                blocklen = 0
                blocknumber += 1
                
            current_block.instruction_list.append(instruction)
            blocklen += 1

            index += 1
        
        # Add the last block

        current_block.update_length(blocklen)
        self.blocks.append(current_block)

        # Resolve cross-references

        for block in self.blocks:

            jmp_indices = asm.find_opcode_sequence(["JUMP"], block.instruction_list)
            jmp_indices += asm.find_opcode_sequence(["JUMPI"], block.instruction_list)

            for i in jmp_indices:
                try:
                    dest_hex = block.instruction_list[i - 1]['argument']
                    dest = int(dest_hex[2:], 16)
                except:
                    continue

                j = 0

                try:
                    while(self.blocks[j].end_addr < dest):
                        j += 1
                except IndexError:
                    continue


                if not (block.id, self.blocks[j].id) in self.xrefs:
                    self.xrefs.append((block.id, self.blocks[j].id))

            # if the last instruction isn't an unconditional jump or halt, also add a reference to the following block

            try:
                if (block.id < len(self.blocks)) and (block.instruction_list[block.length - 1]['opcode'] not in ['JUMP', 'STOP', 'THROW', 'REVERT', 'INVALID']):
                    if not (block.id, self.blocks[j].id) in self.xrefs:
                        self.xrefs.append((block.id, block.id + 1))
            except UnboundLocalError:
                # quickfix
                continue

    def get_easm(self):

        easm = asm.instruction_list_to_easm(self.instruction_list[0:self.blocks[0].length])

        for block in self.blocks[1:]:
            easm += block.get_easm()

        return easm


from z3 import *
import logging
import re
from laser.ethereum import utils


def solve_path(svm, path, caller = None, owner = None, owner_storage_index = None):

    s = Solver() 

    if(caller is not None):
        s.add(svm.env['caller'] == caller)
    if(owner is not None):
        s.add(svm.storage[owner_storage_index] == owner)

    for edge in path:
        if edge.condition is not None:
            s.add(edge.condition)

    if (s.check() == sat):
        return s.model()

    else:
        return None
        

def execute(svm):

    for loc in svm.send_eth_locs:

        node_addr = loc['address']
        function_name = loc['function_name']

        logging.info("Checking node at " + str(node_addr))

        for path in svm.paths[node_addr]:

            path_checked = False

            for edge in path:

                if (edge.condition is not None):

                    cond = str(edge.condition)

                    if 'caller' in cond and 'storage_' in cond:
                        m = re.search(r'storage_(\d+)', cond)

                        if (m):
                            owner_index = int(m.group(1))
                            path_checked = True

                            logging.debug("Constraint on msg.sender: caller == storage_" + str(owner_index))
                            logging.debug("Checking for writes to storage_" + (str(owner_index)))

                            try:

                                for _node_addr in svm.sstor_node_lists[owner_index]:

                                    logging.debug("Checking node " + str(_node_addr))

                                    for _path in svm.paths[_node_addr]:

                                        # Try to solve for caller != owner

                                        m = utils.solve_path(svm, _path, caller=0x1234, owner=0x2345, owner_storage_index=owner_index)

                                        if m is not None:
                                            print("Possible owner overwrite at node " + str(_node_addr) + ", function " + function_name)
                                            print("Input data:")
                                            for d in m.decls():
                                                print("%s = %s" % (d.name(), hex(m[d].as_long())))
                                            print("Ether sent at node: " + str(node_addr))

                                            m = utils.solve_path(svm, svm.paths[node_addr][0])

                                            print("Send Ether input data:")

                                            for d in m.decls():
                                                print("%s = %s" % (d.name(), hex(m[d].as_long())))

                                        else:
                                            print("Unable to satisfy constraints")     

                                        break  

                            except KeyError:
                                logging.debug("No writes found")

            if not path_checked:

                logging.info("Potential unchecked transfer, verifying...")

                m = solve_path(svm, path)
                
                if m is not None:
                    print("Possible unchecked Ether transfer at node " + str(node_addr) + ", function " + function_name)
                    for d in m.decls():
                        print("%s = %s" % (d.name(), hex(m[d].as_long())))
                else:
                    print("Unable to satisfy constraints")                   
